import functools
import jax
import jax.numpy as jnp

from typing import Sequence


@functools.partial(jax.jit, static_argnames=["reg"])
def fit_gaussian2data(x: jax.Array, reg: float = 1e-5) -> Sequence[jax.Array]:
    """
    Fit a Gaussian model to each time step n given the samples from x as array of shape (b, n, d).

    Parameters
    ----------
    x: chex.Array
        Samples from x as array of shape (b, n, d).
    reg:
        Regularization term to avoid singular covariance matrices.

    Returns
    -------
    mean: chex.Array
        Mean of the Gaussian model as array of shape (n, d).
    tri_lower_matrix: chex.Array
        Lower triangular matrix of the Cholesky decomposition of the covariance matrix as array of shape (n, d, d).
    """
    """ Fit a Gaussian model to each time step given the samples from x """
    b, t, d = x.shape

    mean = jnp.mean(x, axis=0, keepdims=True)

    x_centered = x - mean
    covariance_matrix = jnp.einsum("...btn, ...btm->...tnm", x_centered, x_centered) / (
        b - 1 + 1e-5
    )
    covariance_matrix = covariance_matrix + reg * jnp.eye(d)

    tri_lower_matrix = jax.vmap(jnp.linalg.cholesky)(covariance_matrix)

    return mean[0], tri_lower_matrix



from typing import Dict
from pathlib import Path

import pytest
import jax
import numpy as np
import jax.numpy as jnp

import metrx
from metrx import DistanceMeasures, StatisticalMeasures


CONFIG = {
    "seed": 0,
    "dim": 2,
    "x": {
        "batch_size": 32,
        "time_steps": 64,
        "amplitude": 1.0,
        "frequency": 2.0,
        "offset_x": 0.0,
        "offset_y": 0.0,
        "phase_shift": 0.0,
        "sigma": 0.05,
        "rotation": 0.0,
    },
    "y": {
        "batch_size": 32,
        "time_steps": 64,
        "amplitude": 1.0,
        "frequency": 2.0,
        "offset_x": 0.0,
        "offset_y": 0.0,
        "phase_shift": 1.0,
        "sigma": 0.05,
        "rotation": 0.0,
    },
}
TEST_DIR_PATH = Path(metrx.__file__).parent.parent / "tests"


def _generate_trajectory(
    time: jax.Array,
    amplitude: jax.Array,
    frequency: jax.Array,
    offset_x: jax.Array,
    offset_y: jax.Array,
    phase_shift: jax.Array,
    angle: float = 0.0,
) -> jax.Array:
    """
    Generate a trajectory given the parameters.

    Parameters
    ----------
    time: jax.Array
        Time steps as (1, t) array
    amplitude: jax.Array
        Amplitude of the trajectory as (b, 1) array
    frequency: jax.Array
        Frequency of the trajectory as (b, 1) array
    offset_x: jax.Array
        Offset in x-direction as (b, 1) array
    offset_y: jax.Array
        Offset in y-direction as (b, 1) array
    phase_shift: jax.Array
        Phase shift of the trajectory as (b, 1) array
    angle:
        Rotation angle in radians

    Returns
    -------
    jax.Array
        Trajectory as (b, t, 2) array

    """

    def _rotate(array: jax.Array) -> jax.Array:
        rotation_matrix = jnp.array(
            [[jnp.cos(angle), -jnp.sin(angle)], [jnp.sin(angle), jnp.cos(angle)]]
        )
        return jnp.dot(array, rotation_matrix)

    tau = jnp.stack(
        (
            (time - jnp.max(time) / 2.0) + offset_x,
            amplitude
            * jnp.sin(
                2 * jnp.pi * frequency * (time - jnp.max(time) / 2.0) + phase_shift
            )
            + offset_y,
        ),
        axis=-1,
    )

    return _rotate(tau)


def get_samples(rng_key: jax.Array, **kwargs) -> jax.Array:
    """
    Generate batches of trajectory samples from a common base distribution.

    Parameters
    ----------
    rng_key : jax.Array
        Random number generator key.
    **kwargs : Dict
        Dictionary of trajectory parameters

    Returns
    -------
    jax.Array
        Samples from a specific trajectory distribution.
    """
    time_steps = jnp.linspace(0.0, 1, kwargs["time_steps"])[jnp.newaxis, :]

    rng_key, *rng_samples = jax.random.split(rng_key, num=6)
    amplitudes = (
        kwargs["amplitude"]
        + jax.random.normal(rng_samples[0], shape=(kwargs["batch_size"], 1))
        * kwargs["sigma"]
    )
    frequencies = (
        kwargs["frequency"]
        + jax.random.normal(rng_samples[1], shape=(kwargs["batch_size"], 1))
        * kwargs["sigma"]
    )
    offsets_x = (
        kwargs["offset_x"]
        + jax.random.normal(rng_samples[2], shape=(kwargs["batch_size"], 1))
        * kwargs["sigma"]
    )
    offsets_y = (
        kwargs["offset_y"]
        + jax.random.normal(rng_samples[3], shape=(kwargs["batch_size"], 1))
        * kwargs["sigma"]
    )
    phase_shift = (
        kwargs["phase_shift"]
        + jax.random.normal(rng_samples[4], shape=(kwargs["batch_size"], 1))
        * kwargs["sigma"]
    )

    return _generate_trajectory(
        time_steps,
        amplitudes,
        frequencies,
        offsets_x,
        offsets_y,
        phase_shift,
        kwargs["rotation"],
    )


@pytest.mark.parametrize(
    "dist_type,name",
    [
        (dist_type, name)
        for dist_type, name_list in [
            (DistanceMeasures, DistanceMeasures.list_all_names()),
            (StatisticalMeasures, StatisticalMeasures.list_all_names()),
        ]
        for name in name_list
    ],
)
def test_distances(dist_type: DistanceMeasures | StatisticalMeasures, name: str):

    # set Jax-backend to CPU
    jax.config.update("jax_platform_name", "cpu")
    print(f"Jax backend device: {jax.default_backend()} \n")

    rng_key = jax.random.PRNGKey(CONFIG["seed"])

    rng_key, rng_key_x, rng_key_y = jax.random.split(rng_key, num=3)
    x = get_samples(rng_key_x, **CONFIG["x"])
    y = get_samples(rng_key_y, **CONFIG["y"])
    x_1D, y_1D = jnp.squeeze(x[:, 1, 1]), jnp.squeeze(y[:, 1, 1])

    _measure = dist_type.create_instance(name)
    inputs = (x_1D, y_1D) if name == "CosineDistance" else (x, y)

    if isinstance(_measure, DistanceMeasures):
        costs = jax.vmap(jax.vmap(_measure, in_axes=(None, 0)), in_axes=(0, None))(
            *inputs
        )
        costs_jitted = jax.jit(
            jax.vmap(jax.vmap(_measure, in_axes=(None, 0)), in_axes=(0, None))
        )(*inputs)
    else:
        costs = _measure(*inputs)
        costs_jitted = jax.jit(_measure)(*inputs)

    data = dict(mean=np.mean(costs), std=np.std(costs), median=np.median(costs))
    data_jitted = dict(
        mean=np.mean(costs_jitted),
        std=np.std(costs_jitted),
        median=np.median(costs_jitted),
    )

    # load the results
    loaded = np.load(TEST_DIR_PATH / f"test_datasets/{name}.npz")

    # assert close non-jitted
    assert np.allclose(data["mean"], loaded["mean"]), f"{name} failed: Mean not close"
    assert np.allclose(data["std"], loaded["std"]), f"{name} failed: Std not close"
    assert np.allclose(
        data["median"], loaded["median"]
    ), f"{name} failed: Median not close"

    if name != "WassersteinDistance":
        # todo: find a way to verify closeness for WassersteinDistance as well.
        # assert close jitted
        assert np.allclose(
            data_jitted["mean"], loaded["mean"]
        ), f"{name} failed: Mean not close"
        assert np.allclose(
            data_jitted["std"], loaded["std"]
        ), f"{name} failed: Std not close"
        assert np.allclose(
            data_jitted["median"], loaded["median"]
        ), f"{name} failed: Median not"



from abc import ABC, abstractmethod
from typing import Sequence, Optional, Any, Dict, Tuple, Union

import distrax
import jax
import jax.numpy as jnp
from flax import struct
from ott.geometry import pointcloud
from ott.solvers.linear import sinkhorn
from ott.problems.linear import linear_problem
from ott.geometry.costs import CostFn

from metrx.distance_measures import DistanceMeasures
from metrx.utils import fit_gaussian2data


@struct.dataclass
class StatisticalMeasures(ABC):
    """
    Base class for all statistical measures. This class is used to create a registry of all statistical measures.
    To create an instance of a distance measure, use the `create_instance` method.

    Parameters
    ----------
    _registry: `dict`, class attribute, default = {}.
        A dictionary to store all the statistical measures.

    Returns
    -------
    `DistanceMeasures`
        An instance of the base class for all statistical measures.
    """

    _registry = {}

    def __init_subclass__(cls, **kwargs: Dict) -> None:
        """
        Register all the statistical measures in the registry.

        Parameters
        ----------
        kwargs: `dict`
            A dictionary of keyword arguments.

        Returns
        -------
        'None'
        """
        super().__init_subclass__(**kwargs)
        cls._registry[cls.__name__] = cls

    @classmethod
    def create_instance(
        cls, name: Optional[str] = None, *args: Any, **kwargs: Any
    ) -> "StatisticalMeasures":
        """
        Create an instance of the statistical measure using the name of the statistical measure.

        Parameters
        ----------
        name: `str`, default = None
            The name of the statistical measure.
        args: `Any`
            Arguments passed to the constructor of the statistical measure.
        kwargs: `Any`
            Keyword arguments passed to the constructor of the statistical measure.

        Returns
        -------
        `StatisticalMeasures`
            An instance of the statistical measure.

        """
        if name in cls._registry:
            return cls._registry[name].create(*args, **kwargs)
        else:
            registered = ", ".join([key for key in cls._registry.keys()])
            raise ValueError(
                f"Unknown class name: {name}. Registered measures: {registered}"
            )

    @classmethod
    @abstractmethod
    def construct(cls, *args, **kwargs) -> "StatisticalMeasures":
        """Create an Instance of the statistical distance measure"""

    @classmethod
    def create(cls, *args: Any, **kwargs: Any) -> "StatisticalMeasures":
        """
        Create an instance of the the statistical distance measure.

        Parameters
        ----------
        args: `Any`
            Arguments passed to the constructor of the statistical distance measure.
        kwargs: `Any`
            Keyword arguments passed to the constructor of the statistical distance measure.

        Returns
        -------
        `Statistical distance`
            An instance of the relative entropy divergence measure.
        """
        return cls.construct(*args, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> jax.Array:
        """
        Call the statistical measure.

        Parameters
        ----------
        args: `Any`
            The arguments to pass to the run method.
        kwargs: `Any`
            The keyword arguments to pass to the run method.

        Returns
        -------
        `jax.Array`
            The estimated statistical measure.
        """
        return self.run(*args, **kwargs)

    @abstractmethod
    def run(self, *args: Any, **kwargs: Any) -> jax.Array:
        """
        Estimate the statistical measure.

        Parameters
        ----------
        args: `Any`
            The arguments to pass to the method.
        kwargs: `Any`
            The keyword arguments to pass to the method.

        Returns
        -------
        `jax.Array`
            The estimated statistical measure.
        """
        raise NotImplementedError

    @classmethod
    def list_all_names(cls):
        """
        Get the names of the registered distance measures.

        Returns
        -------
        `List[str]`
            The names of the registered distance measures.
        """
        return list(cls._registry.keys())

    @classmethod
    def list_all(cls):
        """
        Get the classes of all the registered distance measures.

        Returns
        -------
        `List[str]`
            The classes of the registered distance measures.

        """
        return list(cls._registry.values())


# --------------------------------------------------------------------------------------------------------------------
# -------------------------------------------- Relative Entropy Divergence -------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class RelativeEntropy(StatisticalMeasures):
    """
    Compute the relative entropy divergence between two time series data both represented as Gaussian distributions
    for each time step n. The relative entropy divergence is computed as the Kullback-Leibler divergence between the
    two Gaussian distributions for each time step n. Thus, the relative entropy divergence is a measure of time series
    data with equal length.

    Parameters
    ----------
    reverse: `bool`, default = False
        A boolean flag to compute the reverse KL divergence.
    reg: `float`, default = 1e-5
        A float value to control the regularization term to avoid singular covariance matrices.
    mean: `bool`, default = False
        A boolean flag to compute the mean of the relative entropy divergence over all time steps.
    median: `bool`, default = False
        A boolean flag to compute the median of the relative entropy divergence over all time steps.
    total_sum: `bool`, default = False
        A boolean flag to compute the total sum of the relative entropy divergence over all time steps.

    Returns
    -------
    `RelativeEntropy`
        An instance of the relative entropy divergence measure.
    """

    reverse: Optional[bool] = struct.field(default=None, pytree_node=False)
    alpha: Optional[float] = struct.field(default=None, pytree_node=False)
    reg: Optional[float] = struct.field(default=None, pytree_node=False)
    mean: Optional[bool] = struct.field(default=None, pytree_node=False)
    median: Optional[bool] = struct.field(default=None, pytree_node=False)
    total_sum: Optional[bool] = struct.field(default=None, pytree_node=False)

    @classmethod
    def construct(
        cls,
        reverse: bool = False,
        reg: float = 1e-5,
        mean: bool = False,
        median: bool = False,
        total_sum: bool = False,
    ) -> "RelativeEntropy":
        """
        Construct an instance of the relative entropy divergence measure.

        Parameters
        ----------
        reverse: `bool`, default = False
            A boolean flag to compute the reverse KL divergence.
        reg: `float`, default = 1e-5
            A float value to control the regularization term to avoid singular covariance matrices.
        mean: `bool`, default = False
            A boolean flag to compute the mean of the relative entropy divergence over all time steps.
        median: `bool`, default = False
            A boolean flag to compute the median of the relative entropy divergence over all time steps.
        total_sum: `bool`, default = True
            A boolean flag to compute the total sum of the relative entropy divergence over all time steps.

        Returns
        -------
        `RelativeEntropy`
            An instance of the relative entropy divergence measure.
        """
        if not mean and not median and not total_sum:
            total_sum = True
        return cls(
            reverse=reverse, reg=reg, mean=mean, median=median, total_sum=total_sum
        )

    def run(self, x: jax.Array, y: jax.Array) -> jax.Array:
        """
        Run the relative entropy divergence measure.

        Parameters
        ----------
        x: `jax.Array`
            Empirical data of shape (b_x, d) if particles are considered. If time series data is considered, the shape
            is (b_x, n, d).
        y: `jax.Array`
            Empirical data of shape (b_y, d) if particles are considered. If time series data is considered, the shape
            is (b_x, n, d).

        Returns
        -------
        jax.Array
            The relative entropy divergence measure of shape (1, ).
        """
        assert (
            x.shape[1:] == y.shape[1:]
        ), f"The objects inside both batches need to have the same shape. Got {x.shape[1:]} and {y.shape[1:]}."
        assert 3 >= x.ndim >= 1 and 3 >= y.ndim >= 1, (
            f"The two batches need to be of shape (b, d, ) if particles, or (b, n, d) if time series data. "
            f"Got x = {x.shape} and y = {y.shape}."
        )
        if x.ndim == 2:
            x = x[..., jnp.newaxis, :]
        if y.ndim == 2:
            y = y[..., jnp.newaxis, :]

        mean_x, tri_lower_x = fit_gaussian2data(x, self.reg)
        mean_y, tri_lower_y = fit_gaussian2data(y, self.reg)

        mvnrml_x = distrax.MultivariateNormalTri(loc=mean_x, scale_tri=tri_lower_x)
        mvnrml_y = distrax.MultivariateNormalTri(loc=mean_y, scale_tri=tri_lower_y)

        if self.reverse:
            relative_entropy = mvnrml_y.kl_divergence(mvnrml_x)[jnp.newaxis, ...]
        else:
            relative_entropy = mvnrml_x.kl_divergence(mvnrml_y)[jnp.newaxis, ...]

        if self.mean:
            return jnp.mean(relative_entropy, axis=-1)
        elif self.median:
            return jnp.median(relative_entropy, axis=-1)
        return jnp.sum(relative_entropy, axis=-1)


# --------------------------------------------------------------------------------------------------------------------
# -------------------------------------------- Frechet Inception Distance --------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class FrechetInceptionDistance(StatisticalMeasures):
    """
    Compute the Frechet Inception Distance (FID) between two time series data both represented as Gaussian distributions
    for each time step n. The Frechet Inception Distance is a measure of the similarity between two time series data
    distributions. The FID corresponds to the 2-Wasserstein distance between the two Gaussian distributions for each
    time step n.

    Parameters
    ----------
    alpha: `float`, default = 1.0
        A float value to control the proportion of the Covariance matrices. If 0.0, the resulting distance is equivalent
        to the the Euclidean distance between the means of the two Gaussian distributions. If 1.0, the resulting
        distance corresponds to the FID.
    reg: `float`, default = 1e-5
        A float value to control the regularization term to avoid singular covariance matrices.
    mean: `bool`, default = False
        A boolean flag to compute the mean of the FID over all time steps.
    median: `bool`, default = False
        A boolean flag to compute the median of the FID over all time steps.
    total_sum: `bool`, default = True
        A boolean flag to compute the total sum of the FID over all time steps.

    """

    alpha: Optional[float] = struct.field(default=None, pytree_node=False)
    reg: Optional[float] = struct.field(default=None, pytree_node=False)
    mean: Optional[bool] = struct.field(default=None, pytree_node=False)
    median: Optional[bool] = struct.field(default=None, pytree_node=False)
    total_sum: Optional[bool] = struct.field(default=None, pytree_node=False)

    @classmethod
    def construct(
        cls,
        alpha: float = 1.0,
        reg: float = 1e-5,
        mean: bool = False,
        median: bool = False,
        total_sum: bool = False,
    ) -> "FrechetInceptionDistance":
        """
        Construct an instance of the Frechet Inception Distance measure.

        Parameters
        ----------
        alpha: `float`, default = 1.0
            A float value to control the proportion of the Covariance matrices. If 0.0, the resulting distance is
            equivalent to the the Euclidean distance between the means of the two Gaussian distributions. If 1.0, the
            resulting distance corresponds to the FID.
        reg: `float`, default = 1e-5
            A float value to control the regularization term to avoid singular covariance matrices.
        mean: `bool`, default = False
            A boolean flag to compute the mean of the FID over all time steps.
        median: `bool`, default = False
            A boolean flag to compute the median of the FID over all time steps.
        total_sum: `bool`, default = True
            A boolean flag to compute the total sum of the FID over all time steps.

        Returns
        -------
        `FrechetInceptionDistance`
            An instance of the Frechet Inception Distance measure.
        """
        if not mean and not median and not total_sum:
            total_sum = True
        return cls(alpha=alpha, reg=reg, mean=mean, median=median, total_sum=total_sum)

    def run(self, x: jax.Array, y: jax.Array) -> jax.Array:
        """
        Run the total Frechet Inception Distance measure.

        Parameters
        ----------
        x: `jax.Array`
            Empirical data of shape (b_x, d) if particles are considered. If time series data is considered, the shape
            is (b_x, n, d).
        y: `jax.Array`
            Empirical data of shape (b_y, d) if particles are considered. If time series data is considered, the shape
            is (b_x, n, d).

        Returns
        -------
        `jax.Array`
            The Frechet Inception Distance measure of shape (1, ).
        """
        assert (
            x.shape[1:] == y.shape[1:]
        ), f"The objects inside both batches need to have the same shape. Got {x.shape[1:]} and {y.shape[1:]}."
        assert 3 >= x.ndim >= 1 and 3 >= y.ndim >= 1, (
            f"The two batches need to be of shape (b, d, ) if particles, or (b, n, d) if time series data. "
            f"Got x = {x.shape} and y = {y.shape}."
        )
        if x.ndim == 2:
            x = x[..., jnp.newaxis, :]
        if y.ndim == 2:
            y = y[..., jnp.newaxis, :]

        mean_x, tri_lower_x = fit_gaussian2data(x, self.reg)
        mean_y, tri_lower_y = fit_gaussian2data(y, self.reg)

        delta_mean = mean_x - mean_y
        mean_diff = jnp.linalg.norm(delta_mean, axis=-1) ** 2

        cov_x = jnp.einsum("...mi, ...ni -> ...mn", tri_lower_x, tri_lower_x)
        trace_cov_x = jax.vmap(jnp.trace)(cov_x)

        cov_y = jnp.einsum("...mi, ...ni -> ...mn", tri_lower_y, tri_lower_y)
        trace_cov_y = jax.vmap(jnp.trace)(cov_y)

        cov_product = jnp.einsum("...mi, ...ni -> ...mn", cov_x, cov_y)
        eig_vals, eig_vecs = jax.vmap(jnp.linalg.eigh)(cov_product)
        sqrt_cov_product = jnp.einsum(
            "...mj,...j, ...nj->...mn", eig_vecs, jnp.sqrt(eig_vals), eig_vecs
        )
        trace_sqrt_product = jax.vmap(jnp.trace)(sqrt_cov_product)

        distances = mean_diff + self.alpha * (
            trace_cov_x + trace_cov_y - 2 * trace_sqrt_product
        )

        if self.mean:
            return jnp.mean(distances, axis=-1)
        elif self.median:
            return jnp.median(distances, axis=-1)
        return jnp.sum(distances, axis=-1)


# --------------------------------------------------------------------------------------------------------------------
# --------------------------------------------- Maximum Mean Discrepancy ---------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class MaximumMeanDiscrepancy(StatisticalMeasures):
    """
    Compute the Maximum Mean Discrepancy (MMD) between two time series data. Depending on the selected distance
    the two time series have to be of the same length (N_x = N_y), .e.g., the squared Euclidean distance. However,
    using dynamic time warping as distance measure, the two time series can have different lengths (N_x != N_y).

    Parameters
    ----------
    distance: `DistanceMeasures`, default = SquaredEuclideanDistance
        A distance measure to compute the distance between the two time series data.
    bandwidths: `Sequence[float]`, default = [1.0, 10.0, 20.0, 40.0, 80.0, 100.0, 130.0, 200.0, 400.0, 800.0, 1000.0]
        A sequence of bandwidths to compute the MMD.
    unbiased: `bool`, default = True
        A boolean flag to compute the unbiased MMD.

    Returns
    -------
    `MaximumMeanDiscrepancy`
        An instance of the Maximum Mean Discrepancy measure
    """

    distance: DistanceMeasures = struct.field(pytree_node=False)

    bandwidths: Tuple = struct.field(pytree_node=False)
    unbiased: Optional[bool] = struct.field(default=True, pytree_node=False)

    @classmethod
    def construct(
        cls,
        distance_measure: Union[DistanceMeasures, str] = "EuclideanDistance",
        distance_kwargs: Dict = {},
        bandwidths: Optional[Union[Sequence[float], Tuple]] = None,
        unbiased: bool = True,
    ) -> "MaximumMeanDiscrepancy":
        """
        Construct an instance of the Maximum Mean Discrepancy measure.

        Parameters
        ----------
        distance: `DistanceMeasures`, default = None
            A distance measure to compute the distance between the two time series data. If None, the Squared Euclidean
            distance is used.
        bandwidths: `Sequence[float]`, default = None
            A sequence of bandwidths to compute the MMD. If None, the default bandwidths are used.
        unbiased
            A boolean flag to compute the unbiased MMD. If None, the unbiased MMD is computed.

        Returns
        -------
        `MaximumMeanDiscrepancy`
            An instance of the Maximum Mean Discrepancy measure
        """
        if isinstance(distance_measure, str):
            distance_measure = DistanceMeasures.create_instance(
                distance_measure, **distance_kwargs
            )

        if bandwidths is None:
            bandwidths = (
                1.0,
                10.0,
                20.0,
                40.0,
                80.0,
                100.0,
                130.0,
                200.0,
                400.0,
                800.0,
                1000.0,
            )

        elif isinstance(bandwidths, Sequence):
            bandwidths = tuple(bandwidths)

        return cls(distance=distance_measure, bandwidths=bandwidths, unbiased=unbiased)

    def _mmd_kernel(self, x: jax.Array, y: jax.Array) -> jax.Array:
        """
        Compute the Maximum Mean Discrepancy kernel between two time series data. Depending on the selected distance
        the two time series have to be of the same length (N_x = N_y), .e.g., the squared Euclidean distance. However,
        using dynamic time warping as distance measure, the two time series can have different lengths (N_x != N_y).

        Parameters
        ----------
        x: `jax.Array`
            Empirical data of shape (B_x, N_x, D).
        y: `jax.Array`
            Empirical data of shape (B_y, N_y, D).

        Returns
        -------
        `jax.Array`
            The Maximum Mean Discrepancy kernel of shape (B_x, B_y).
        """
        distance_matrix = jax.vmap(
            jax.vmap(self.distance, in_axes=(None, 0)), in_axes=(0, None)
        )(x, y)

        def rbf_kernel(kernelized_dist: jax.Array, bandwidth: float) -> Any:
            return kernelized_dist + jnp.exp(-0.5 / bandwidth * distance_matrix), None

        kernelized_distance_matrix = jnp.zeros((x.shape[0], y.shape[0]))
        kernelized_distance_matrix, _ = jax.lax.scan(
            rbf_kernel, kernelized_distance_matrix, jnp.array(self.bandwidths)
        )
        return kernelized_distance_matrix

    def run(self, x: jax.Array, y: jax.Array) -> jax.Array:
        """
        Run the Maximum Mean Discrepancy measure.

        Parameters
        ----------
        x: `jax.Array`
            Empirical data of shape (b_x, d) if particles are considered. If time series data is considered, the shape
            is (b_x, n_x, d).
        y: `jax.Array`
            Empirical data of shape (b_y, d) if particles are considered. If time series data is considered, the shape
            is (b_x, n_y, d).

        Returns
        -------
        `jax.Array`
            The Maximum Mean Discrepancy measure of shape ().
        """
        assert x.shape[-1] == y.shape[-1], (
            f"The objects inside both batches need to have the same dimensionality. "
            f"Got {x.shape[1:]} and {y.shape[1:]}."
        )
        assert 3 >= x.ndim >= 1 and 3 >= y.ndim >= 1, (
            f"The two batches need to be of shape (b, d, ) if particles, or (b, n, d) if time series data. "
            f"Got x = {x.shape} and y = {y.shape}."
        )
        if x.ndim == 2:
            x = x[..., jnp.newaxis, :]
        if y.ndim == 2:
            y = y[..., jnp.newaxis, :]

        kxx = self._mmd_kernel(x, x)
        i, j = jnp.diag_indices(kxx.shape[-1])
        kxx = kxx.at[..., i, j].set(0.0)

        kyy = self._mmd_kernel(y, y)
        i, j = jnp.diag_indices(kyy.shape[-1])
        kyy = kyy.at[..., i, j].set(0.0)

        kxy = self._mmd_kernel(x, y)

        b_x, b_y = x.shape[0], y.shape[0]
        c_xy = 2.0 / (b_x * b_y)
        if self.unbiased:
            c_xx = 1 / (b_x * (b_x - 1)) if b_x > 1 else 1 / b_x
            c_yy = 1 / (b_y * (b_y - 1)) if b_y > 1 else 1 / b_y
        else:
            c_xx = 1 / b_x
            c_yy = 1 / b_y

        return c_xx * jnp.sum(kxx) - c_xy * jnp.sum(kxy) + c_yy * jnp.sum(kyy)


# --------------------------------------------------------------------------------------------------------------------
# ---------------------------------------------- 1-Wasserstein distance ----------------------------------------------
# --------------------------------------------------------------------------------------------------------------------


@struct.dataclass
class OTTCostWrapper(CostFn):
    distance_measure: DistanceMeasures = struct.field(pytree_node=False)

    @classmethod
    def construct(cls, distance_measure: DistanceMeasures):
        return cls(distance_measure)

    def __call__(self, x: jax.Array, y: jax.Array) -> float:
        distance = self.distance_measure(x, y)
        return distance.squeeze()


@struct.dataclass
class WassersteinDistance(StatisticalMeasures):
    """
    Approximation of the 1-Wasserstein distance using the Sinkhorn iterations.
    Depending on the selected distance the two time series have to be of the same length (N_x = N_y), .e.g., the squared Euclidean distance. However,
    using dynamic time warping as distance measure, the two time series can have different lengths (N_x != N_y).

    Parameters
    ----------
    distance: `DistanceMeasures`, default = SquaredEuclideanDistance
        A distance measure to compute the distance between the two time series data.
    distance_kwargs: `Dict`, containing the configurations to intitialize the `DistanceMeasures`, if `DistanceMeasurs` is a str.
    epsilon: `float`, default = None
        The regularization parameter for the regularized optimal transport problem
    return_regularized_cost: `bool`, optional, default = None.
        If True, the regularized cost is returned. If False, the unregularized cost is returned.

    Returns
    -------
    `WassersteinDistance`
        An instance of the Wasserstein distance measure
    """

    cost_fn: OTTCostWrapper = struct.field(pytree_node=False)
    epsilon: Optional[float] = struct.field(default=None, pytree_node=False)
    return_regularized_cost: bool = struct.field(default=False, pytree_node=False)

    @classmethod
    def construct(
        cls,
        distance_measure: Union[DistanceMeasures, str] = "EuclideanDistance",
        distance_kwargs: Dict = {},
        epsilon: Optional[float] = None,
        return_regularized_cost: bool = False,
    ) -> StatisticalMeasures:

        if isinstance(distance_measure, str):
            distance_measure = DistanceMeasures.create_instance(
                distance_measure, **distance_kwargs
            )

        # Build cost function wrapper for the ott-jax library
        cost_fn = OTTCostWrapper.construct(distance_measure)

        return cls(cost_fn, epsilon, return_regularized_cost)

    @property
    def distance_measure(self):
        return self.cost_fn.distance_measure

    def run(self, x: jax.Array, y: jax.Array) -> jax.Array:
        """
        Run the 1-Wasserstein distance.

        Parameters
        ----------
        x: `jax.Array`
            Empirical data of shape (b_x, d) if particles are considered. If time series data is considered, the shape
            is (b_x, n_x, d).
        y: `jax.Array`
            Empirical data of shape (b_y, d) if particles are considered. If time series data is considered, the shape
            is (b_x, n_y, d).

        Returns
        -------
        `jax.Array`
            The Maximum Mean Discrepancy measure of shape ().
        """
        geom = pointcloud.PointCloud(x, y, cost_fn=self.cost_fn, epsilon=self.epsilon)
        ot_prob = linear_problem.LinearProblem(geom)
        solver = sinkhorn.Sinkhorn()
        out = solver(ot_prob)

        if self.return_regularized_cost:
            return out.reg_ot_cost

        return jnp.sum(out.matrix * out.geom.cost_matrix)



import jax
import jax.numpy as jnp

from abc import ABC, abstractmethod
from flax import struct
from ott.geometry import pointcloud
from ott.solvers.linear import sinkhorn
from ott.problems.linear import linear_problem
from ott.geometry.costs import CostFn
from typing import Sequence, Optional, Any, Dict


@struct.dataclass
class DistanceMeasures(ABC):
    """
    Base class for all distance measures. The distance measures are used to calculate the similarity between two data
    points. This points could be either one dimensional points or time-series data. The base class provides a registry
    to store all the implemented distance measures. To create an instance of a distance measure, use
    the `create_instance` method.

    Parameters
    ----------
    _registry: `dict`, class attribute, default = {}.
        A dictionary to store all the distance measures.

    Returns
    -------
    `DistanceMeasures`
        An instance of the base class for all distance measures.
    """

    _registry = {}

    def __init_subclass__(cls, **kwargs: Dict) -> None:
        """
        Initialize the subclass of the DistanceMeasures class.

        Parameters
        ----------
        kwargs: `Dict`
            The keyword arguments to pass to the superclass.

        Returns
        -------
        `None`
        """
        super().__init_subclass__(**kwargs)
        cls._registry[cls.__name__] = cls

    @classmethod
    def create_instance(
        cls, name: Optional[str] = None, *args: Any, **kwargs: Any
    ) -> "DistanceMeasures":
        """
        Create an instance of the distance measure.

        Parameters
        ----------
        name: `str`, optional, default = None.
            The name of the distance measure to create.
        args: `Any`
            The arguments to pass to the constructor of the distance measure.
        kwargs: `Any`
            The keyword arguments to pass to the constructor of the distance measure.

        Returns
        -------
        `DistanceMeasures`
            An instance of the distance measure.
        """
        if name in cls._registry:
            return cls._registry[name].create(*args, **kwargs)
        else:
            registered = ", ".join([key for key in cls._registry.keys()])
            raise ValueError(
                f"Unknown class name: {name}. Registered measures: {registered}"
            )

    @classmethod
    @abstractmethod
    def construct(cls, *args, **kwargs) -> "DistanceMeasures":
        """Create an Instance of the Distance Measure"""

    @classmethod
    def create(cls, *args: Any, **kwargs: Any) -> "DistanceMeasures":
        """
        Create an instance of the Minkowski distance measure.

        Parameters
        ----------
        args: `Any`
            The arguments to pass to the constructor.
        kwargs: `Any`
            The keyword arguments to pass to the constructor.

        Returns
        -------
        `DistanceMeasures`:
            An instance of the distance measure.
        """
        return cls.construct(*args, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> jax.Array:
        """
        Call the distance measure.

        Parameters
        ----------
        args: `Any`
            The arguments to pass to the run method.
        kwargs: `Any`
            The keyword arguments to pass to the run method.

        Returns
        -------
        `jax.Array`
            The estimated distance measure.
        """
        return self.run(*args, **kwargs)

    @abstractmethod
    def run(self, *args: Any, **kwargs: Any) -> jax.Array:
        """
        Estimate the distance measure.

        Parameters
        ----------
        args: `Any`
            The arguments to pass to the method.
        kwargs: `Any`
            The keyword arguments to pass to the method.

        Returns
        -------
        `jax.Array`
            The estimated distance measure.
        """
        raise NotImplementedError

    @classmethod
    def list_all_names(cls):
        """
        Get the names of the registered distance measures.

        Returns
        -------
        `List[str]`
            The names of the registered distance measures.
        """
        return list(cls._registry.keys())

    @classmethod
    def list_all(cls):
        """
        Get the classes of all the registered distance measures.

        Returns
        -------
        `List[str]`
            The classes of the registered distance measures.

        """
        return list(cls._registry.values())


# --------------------------------------------------------------------------------------------------------------------
# ------------------------------------------------ Minkowski Distance ------------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class MinkowskiDistance(DistanceMeasures):
    """
    Similarity measure between data points using the Minkowski distance. The Minkowski distance is a metric in a normed
    vector space. It is the L_p norm of the difference between two vectors. If data point corresponds to a time series,
    the distance is calculated between the individual pairs of points. This implies that the series must be of equal
    length.

    Parameters
    ----------
    p: `int`, optional, default = 2.
        The order of the Minkowski distance.
    mean: `bool`, optional, default = False.
        If True, the mean of the distances is returned.
    median: `bool`, optional, default = False.
        If True, the median of the distances is returned.
    total_sum: `bool`, optional, default = True.
        If True, the total sum of the distances is returned.

    Returns
    -------
    `MinkowskiDistance`
        An instance of the Minkowski distance measure.
    """

    p: Optional[float] = struct.field(default=None, pytree_node=False)
    mean: Optional[bool] = struct.field(default=None, pytree_node=False)
    median: Optional[bool] = struct.field(default=None, pytree_node=False)
    total_sum: Optional[bool] = struct.field(default=None, pytree_node=False)

    @classmethod
    def construct(
        cls,
        p: float = 2,
        mean: bool = False,
        median: bool = False,
        total_sum: bool = False,
    ) -> "MinkowskiDistance":
        """
        Construct the Minkowski distance measure.

        Parameters
        ----------
        p: `int`, default = 2.
            The order of the Minkowski distance.
        mean: `bool`, default = False.
            If True, the mean of the distances is returned.
        median: `bool`, default = False.
            If True, the median of the distances is returned.
        total_sum: `bool`, default = True.
            If True, the total sum of the distances is returned.

        Returns
        -------
        `MinkowskiDistance`
            An instance of the Minkowski distance measure.
        """
        assert (
            p >= 1
        ), "The order p of the Minkowski distance should be in [0, inf]. Got {p}"
        if not mean and not median and not total_sum:
            total_sum = True
        return cls(p=p, mean=mean, median=median, total_sum=total_sum)

    def run(self, x: jax.Array, y: Optional[jax.Array] = None) -> jax.Array:
        """
        Estimate the Minkowski distance measure.

        Parameters
        ----------
        x: `jax.Array`
            The input data of shape (d, ) if particle, or (n, d) if time series.
        y: `jax.Array`, optional
            The second input data of shape (d, ) if particle, or (n, d) if time series. If not provided, y = 0.

        Returns
        -------
        `jax.Array`:
            The estimated Minkowski distance of shape ().
        """
        if y is None:
            y = jnp.zeros_like(x)

        assert (
            x.ndim == y.ndim and x.shape == y.shape
        ), f"The two data points need to be of the same shape. Got x={x.shape} and y={y.shape}."
        assert x.ndim <= 2 and y.ndim <= 2, (
            f"The two inputs need to be of shape (d, ) if particle, or (n, d) if time series. "
            f"Got x = {x.shape} and y = {y.shape}."
        )

        if x.ndim == 1:
            x = jnp.expand_dims(x, axis=0)  # (n = 1, d)
            y = jnp.expand_dims(y, axis=0)  # (n = 1, d)

        distance = jnp.linalg.norm(x - y, axis=-1, ord=self.p)  # (n, )
        if self.mean:
            return jnp.mean(distance, axis=-1)
        elif self.median:
            return jnp.median(distance, axis=-1)
        return jnp.sum(distance, axis=-1)


# --------------------------------------------------------------------------------------------------------------------
# ------------------------------------------------ Euclidean Distance ------------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class EuclideanDistance(DistanceMeasures):
    """
    Similarity measure between data points using the Euclidean distance. The Euclidean distance is a metric in a normed
    vector space. It is the L_p norm of the difference between two vectors. If data point corresponds to a time series,
    the distance is calculated between the individual pairs of points. This implies that the series must be of equal
    length.

    Parameters
    ----------
    mean: `bool`, optional, default = False.
        If True, the mean of the distances is returned.
    median: `bool`, optional, default = False.
        If True, the median of the distances is returned.
    total_sum: `bool`, optional, default = True.
        If True, the total sum of the distances is returned.

    Returns
    -------
    `EuclideanDistance`
        An instance of the Euclidean distance measure.
    """

    mean: Optional[bool] = struct.field(default=None, pytree_node=False)
    median: Optional[bool] = struct.field(default=None, pytree_node=False)
    total_sum: Optional[bool] = struct.field(default=None, pytree_node=False)

    @classmethod
    def construct(
        cls, mean: bool = False, median: bool = False, total_sum: bool = False
    ) -> "EuclideanDistance":
        """
        Construct the Euclidean distance measure.

        Parameters
        ----------
        mean: `bool`, default = False.
            If True, the mean of the distances is returned.
        median: `bool`, default = False.
            If True, the median of the distances is returned.
        total_sum: `bool`, default = True.
            If True, the total sum of the distances is returned.

        Returns
        -------
        `EuclideanDistance`
            An instance of the Euclidean distance measure.
        """
        if not mean and not median and not total_sum:
            total_sum = True
        return cls(mean=mean, median=median, total_sum=total_sum)

    def run(self, x: jax.Array, y: Optional[jax.Array] = None) -> jax.Array:
        """
        Estimate the Euclidean distance measure.

        Parameters
        ----------
        x: `jax.Array`
            The input data point of shape (d, ) if particle, or (n, d) if time series.
        y: `jax.Array`, optional
            The second input data of shape (d, ) if particle, or (n, d) if time series. If not provided, y = 0.

        Returns
        -------
        `jax.Array`:
            The estimated Euclidean distance of shape ().
        """
        if y is None:
            y = jnp.zeros_like(x)

        assert (
            x.ndim == y.ndim and x.shape == y.shape
        ), f"The two data points need to be of the same shape. Got x={x.shape} and y={y.shape}."
        assert x.ndim <= 2 and y.ndim <= 2, (
            f"The two inputs need to be of shape (d, ) if particle, or (n, d) if time series. "
            f"Got x = {x.shape} and y = {y.shape}."
        )

        if x.ndim == 1:
            x = jnp.expand_dims(x, axis=0)  # (n = 1, d)
            y = jnp.expand_dims(y, axis=0)  # (n = 1, d)

        distance = jnp.linalg.norm(x - y, axis=-1)  # (n, )
        if self.mean:
            return jnp.mean(distance, axis=-1)
        elif self.median:
            return jnp.median(distance, axis=-1)
        return jnp.sum(distance, axis=-1)


# --------------------------------------------------------------------------------------------------------------------
# -------------------------------------------- Squared Euclidean Distance --------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class SquaredEuclideanDistance(EuclideanDistance):
    """
    Similarity measure between a data point using the squared Euclidean distance. The distance is not metric in a normed
    vector space as it does not satisfy the triangle inequality. If data point corresponds to a time series,
    the distance is calculated between the individual pairs of points. This implies that the series must be of equal
    length.

    Parameters
    ----------
    mean: `bool`, optional, default = False.
        If True, the mean of the distances is returned.
    median: `bool`, optional, default = False.
        If True, the median of the distances is returned.
    total_sum: `bool`, optional, default = True.
        If True, the total sum of the distances is returned.

    Returns
    -------
    `SquaredEuclideanDistance`
        An instance of the squared Euclidean distance measure.
    """

    def run(self, x: jax.Array, y: Optional[jax.Array] = None) -> jax.Array:
        """
        Estimate the squared Euclidean distance measure.

        Parameters
        ----------
        x: `jax.Array`
            The input data point of shape (d, ) if particle, or (n, d) if time series.
        y: `jax.Array`, optional
            The second input data of shape (d, ) if particle, or (n, d) if time series. If not provided, y = 0.

        Returns
        -------
        `jax.Array`:
            The estimated squared Euclidean distance of shape ().

        """
        if y is None:
            y = jnp.zeros_like(x)

        assert (
            x.ndim == y.ndim and x.shape == y.shape
        ), f"The two data points need to be of the same shape. Got x={x.shape} and y={y.shape}."

        assert x.ndim <= 2 and y.ndim <= 2, (
            f"The two inputs need to be of shape (d, ) if particle, or (n, d) if time series."
            f"Got x = {x.shape} and y = {y.shape}."
        )

        if x.ndim == 1:
            x = jnp.expand_dims(x, axis=0)  # (n = 1, d)
            y = jnp.expand_dims(y, axis=0)  # (n = 1, d)

        squared_distance = jnp.linalg.norm(x - y, axis=-1) ** 2  # (n, )
        if self.mean:
            return jnp.mean(squared_distance, axis=-1)
        elif self.median:
            return jnp.median(squared_distance, axis=-1)
        return jnp.sum(squared_distance, axis=-1)


# --------------------------------------------------------------------------------------------------------------------
# ------------------------------------------------- Cosine Distance --------------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class CosineDistance(DistanceMeasures):
    eps: float = struct.field(default=1e-8, pytree_node=False)

    @classmethod
    def construct(cls, eps: float = 1e-8) -> "CosineDistance":
        return cls(eps)

    def run(self, x: jax.Array, y: jax.Array) -> jax.Array:
        """
        Estimate the Cosine distance measure d = 1 - 'cosine similarity'.
        Note that the cosine distance is only a quasi-metric.

        Parameters
        ----------
        x: `jax.Array`
            The input data point of shape (d, ). Time series data is not supported at this point
        y: `jax.Array`, optional
            The second input data of shape (d, ).

        Returns
        -------
        `jax.Array`:
            The estimated cosine distance of shape ().

        """
        assert (
            x.shape == y.shape
        ), f"The two data points need to be of the same shape. Got x={x.shape} and y={y.shape}."

        # Deal with pseudo-time series data, i.e., data of shape (1, d)
        if x.ndim == 2 and x.shape[0] == 1:
            x = x.squeeze(0)
        if y.ndim == 2 and y.shape[0] == 1:
            y = y.squeeze(0)

        assert (
            x.ndim == y.ndim <= 1
        ), f"The cosine distance only supports data of shape (d,), but received {x.shape}"

        norm_x = jnp.maximum(jnp.linalg.norm(x), self.eps)
        norm_y = jnp.maximum(jnp.linalg.norm(y), self.eps)

        dot_product = jnp.dot(x, y)
        cosine_similarity = dot_product / (norm_x * norm_y)
        # cosine_similarity = jnp.clip(cosine_similarity, -1.0, 1.0)
        return jnp.clip(1 - cosine_similarity, 0.0, 2.0)


# --------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------- Mahalanobis Distance -----------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class MahalanobisDistance(DistanceMeasures):
    """
    Similarity measure between a data point and a multivariate Normal using the Mahalanobis distance.
    If data point corresponds to a time series, the distance is calculated between the individual pairs.
    This implies that the series must be of equal length.

    Parameters
    ----------
    mean: `bool`, optional, default = False.
        If True, the mean of the distances is returned.
    median: `bool`, optional, default = False.
        If True, the median of the distances is returned.
    total_sum: `bool`, optional, default = True.
        If True, the total sum of the distances is returned.

    Returns
    -------
    `MahalanobisDistance`
        An instance of the Mahalanobis distance measure.
    """

    mean: Optional[bool] = struct.field(default=None, pytree_node=False)
    median: Optional[bool] = struct.field(default=None, pytree_node=False)
    total_sum: Optional[bool] = struct.field(default=None, pytree_node=False)

    @classmethod
    def construct(
        cls, mean: bool = False, median: bool = False, total_sum: bool = False
    ) -> "MahalanobisDistance":
        """
        Construct the Mahalanobis distance measure.

        Parameters
        ----------
        mean: `bool`, default = False.
            If True, the mean of the distances is returned.
        median: `bool`, default = False.
            If True, the median of the distances is returned.
        total_sum: `bool`, default = True.
            If True, the total sum of the distances is returned.

        Returns
        -------
        `MahalanobisDistance`
            An instance of the Mahalanobis distance measure.
        """
        if not mean and not median and not total_sum:
            total_sum = True
        return cls(mean=mean, median=median, total_sum=total_sum)

    def run(
        self,
        x: jax.Array,
        mu: Optional[jax.Array] = None,
        covariance_matrix: Optional[jax.Array] = None,
        precision_matrix: Optional[jax.Array] = None,
    ) -> jax.Array:
        """
        Estimate the Mahalanobis distance measure.

        Parameters
        ----------
        x: `jax.Array`
            The input data point of shape (d, ) if particle, or (n, d) if time series.
        mu: `jax.Array`, optional
            The mean data point of shape (d, ) if particle, or (n, d) if time series. If not provided, mu = 0.
        covariance_matrix: `jax.Array`, optional
            The covariance matrix of shape (d, d, ) if particle, or (n, d, d) if time series. If not provided,
            covariance_matrix = eye(d).
        precision_matrix: `jax.Array`, optional
            The precision matrix of shape (d, d, ) if particle, or (n, d, d) if time series. If not provided,
            precision_matrix = eye(d).

        Returns
        -------
        `jax.Array`:
            The estimated Mahalanobis distance of shape ().
        """
        if mu is None:
            mu = jnp.zeros_like(x)

        assert (
            x.ndim == mu.ndim and x.shape == mu.shape
        ), f"The data point and the mean need to be of the same shape. Got x={x.shape}."
        assert x.ndim <= 2 and mu.ndim <= 2, (
            f"The two inputs need to be of shape (d, ) if particle, or (n, d) if time series. "
            f"Got x = {x.shape} and y = {mu.shape}."
        )

        if x.ndim == 1:
            x = jnp.expand_dims(x, axis=0)
            mu = jnp.expand_dims(mu, axis=0)

        lower_tri_covariance = None
        if covariance_matrix is not None:
            assert (
                x.ndim == covariance_matrix.ndim - 1
                and x.shape == covariance_matrix.shape[:-1]
            ), (
                f"The data point and the covariance need to be of the same shape. "
                f"Got x={x.shape} and y={covariance_matrix.shape}."
            )
            if covariance_matrix.ndim == 2:
                covariance_matrix = covariance_matrix[jnp.newaxis, ...]
            lower_tri_covariance = jax.vmap(jnp.linalg.cholesky)(covariance_matrix)

        lower_tri_precision = None
        if precision_matrix is not None:
            assert (
                x.ndim == precision_matrix.ndim - 1
                and x.shape == precision_matrix.shape[:-1]
            ), (
                f"The data point and the covariance need to be of the same shape. "
                f"Got x={x.shape} and y={precision_matrix.shape}."
            )
            if precision_matrix.ndim == 2:
                precision_matrix = precision_matrix[jnp.newaxis, ...]
            lower_tri_precision = jax.vmap(jnp.linalg.cholesky)(precision_matrix)

        if covariance_matrix is None and precision_matrix is None:
            n, d = mu.shape
            lower_tri_precision = jnp.expand_dims(jnp.eye(d), 0).repeat(n, axis=0)

        if lower_tri_covariance is not None:
            x_transformed = jax.vmap(jnp.linalg.solve)(
                lower_tri_covariance, x - mu
            )  # (n, d)
        elif lower_tri_precision is not None:
            x_transformed = jnp.einsum(
                "...mn, ...n->...m", lower_tri_precision, x - mu
            )  # (n, d)
        else:
            raise ValueError("Neither covariance nor precision matrices were given.")

        distance = jnp.linalg.norm(x_transformed, axis=-1)  # (n, )
        if self.mean:
            return jnp.mean(distance, axis=-1)
        elif self.median:
            return jnp.median(distance, axis=-1)
        return jnp.sum(distance, axis=-1)


# --------------------------------------------------------------------------------------------------------------------
# ------------------------------------------- Squared Mahalanobis Distance -------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class SquaredMahalanobisDistance(MahalanobisDistance):
    """
    Similarity measure between a data point and a multivariate Normal using the squared Mahalanobis distance.
    If data point corresponds to a time series, the distance is calculated between the individual pairs.
    This implies that the series must be of equal length.

    Parameters
    ----------
    mean: `bool`, optional, default = False.
        If True, the mean of the distances is returned.
    median: `bool`, optional, default = False.
        If True, the median of the distances is returned.
    total_sum: `bool`, optional, default = True.
        If True, the total sum of the distances is returned.

    Returns
    -------
    `SquaredMahalanobisDistance`
        An instance of the squared Mahalanobis distance measure.
    """

    def run(
        self,
        x: jax.Array,
        mu: Optional[jax.Array] = None,
        covariance_matrix: Optional[jax.Array] = None,
        precision_matrix: Optional[jax.Array] = None,
    ) -> jax.Array:
        """
         Estimate the Mahalanobis distance measure.

         Parameters
         ----------
        x: `jax.Array`
             The input data point of shape (d, ) if particle, or (n, d) if time series.
         mu: `jax.Array`, optional
             The mean data point of shape (d, ) if particle, or (n, d) if time series. If not provided, mu = 0.
         covariance_matrix: `jax.Array`, optional
             The covariance matrix of shape (d, d, ) if particle, or (n, d, d) if time series. If not provided,
             covariance_matrix = eye(d).
         precision_matrix: `jax.Array`, optional
             The precision matrix of shape (d, d, ) if particle, or (n, d, d) if time series. If not provided,
             precision_matrix = eye(d).

         Returns
         -------
         `jax.Array`:
             The estimated squared Mahalanobis distance of shape ().
        """
        if mu is None:
            mu = jnp.zeros_like(x)

        assert (
            x.ndim == mu.ndim and x.shape == mu.shape
        ), f"The data point and the mean need to be of the same shape. Got x={x.shape}."
        assert x.ndim <= 2 and mu.ndim <= 2, (
            f"The two inputs need to be of shape (d, ) if particle, or (n, d) if time series. "
            f"Got x = {x.shape} and y = {mu.shape}."
        )

        if x.ndim == 1:
            x = jnp.expand_dims(x, axis=0)
            mu = jnp.expand_dims(mu, axis=0)

        lower_tri_covariance = None
        if covariance_matrix is not None:
            assert (
                x.ndim == covariance_matrix.ndim - 1
                and x.shape == covariance_matrix.shape[:-1]
            ), (
                f"The data point and the covariance need to be of the same shape. "
                f"Got x={x.shape} and y={covariance_matrix.shape}."
            )
            if covariance_matrix.ndim == 2:
                covariance_matrix = covariance_matrix[jnp.newaxis, ...]
            lower_tri_covariance = jax.vmap(jnp.linalg.cholesky)(covariance_matrix)

        lower_tri_precision = None
        if precision_matrix is not None:
            assert (
                x.ndim == precision_matrix.ndim - 1
                and x.shape == precision_matrix.shape[:-1]
            ), (
                f"The data point and the covariance need to be of the same shape. "
                f"Got x={x.shape} and y={precision_matrix.shape}."
            )
            if precision_matrix.ndim == 2:
                precision_matrix = precision_matrix[jnp.newaxis, ...]
            lower_tri_precision = jax.vmap(jnp.linalg.cholesky)(precision_matrix)

        if covariance_matrix is None and precision_matrix is None:
            n, d = mu.shape
            lower_tri_precision = jnp.expand_dims(jnp.eye(d), 0).repeat(n, axis=0)

        if lower_tri_covariance is not None:
            x_transformed = jax.vmap(jnp.linalg.solve)(
                lower_tri_covariance, x - mu
            )  # (n, d)
        elif lower_tri_precision is not None:
            x_transformed = jnp.einsum(
                "...mn, ...n->...m", lower_tri_precision, x - mu
            )  # (n, d)
        else:
            raise ValueError("Neither covariance nor precision matrices were given.")

        squared_distance = jnp.linalg.norm(x_transformed, axis=-1) ** 2  # (n, )

        if self.mean:
            return jnp.mean(squared_distance, axis=-1)
        elif self.median:
            return jnp.median(squared_distance, axis=-1)
        return jnp.sum(squared_distance, axis=-1)


# --------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------- Dynamic Time Warping -----------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class DynamicTimeWarping(DistanceMeasures):
    """
    Similarity measure between time series data using the Dynamic Time Warping (DTW) [1, 2]. The DTW is an algorithm
    used to measure the similarity between two sequences that may vary in time or speed. This implies that the series
    can be of different lengths. The resulting DTW distance is just a similarity measure and not a metric as it does
    not satisfy the triangle inequality.

    The algorithm follows a dynamic programming approach to find the optimal alignment between the two sequences. We
    used the DTW implementation from [3].

    Parameters
    ----------
    distance: `DistanceMeasures`, optional, default = None.
        The distance measure to use for the DTW. If None, the Euclidean distance is used.

    Returns
    -------
    `DynamicTimeWarping`
        An instance of the Dynamic Time Warping distance measure.

    References
    ----------
    [1] T. K. Vintsyuk. Speech discrimination by dynamic programming. Cybernetics, 4(1):52–57,1968.
        Available: https://link.springer.com/article/10.1007/BF01074755
    [2] H. Sakoe and S. Chiba. Dynamic programming algorithm optimization for spoken word recognition.
        IEEE transactions on acoustics, speech, and signal processing, 26(1):43–49, 1978.
        Available: https://ieeexplore.ieee.org/document/1163055
    [3] K. Heidler. (Soft-)DTW for JAX, Github, https://github.com/khdlr/softdtw_jax
    """

    distance: Optional[DistanceMeasures] = struct.field(default=None, pytree_node=False)

    @classmethod
    def construct(
        cls, distance: Optional[DistanceMeasures] = None
    ) -> "DynamicTimeWarping":
        """
        Construct the Dynamic Time Warping distance measure.

        Parameters
        ----------
        distance: `DistanceMeasures`, optional, default = None.
            The distance measure to use for the DTW. If None, the Euclidean distance is used.

        Returns
        -------
        `DynamicTimeWarping`
            An instance of the Dynamic Time Warping distance measure.
        """
        if distance is None:
            distance = EuclideanDistance.construct()
        return cls(distance=distance)

    def init_model_matrix(self, x: jax.Array, y: jax.Array) -> jax.Array:
        """
        Initialize the state for the Dynamic Time Warping distance measure.

        Parameters
        ----------
        x: `jax.Array`
            The input data of shape = (n_x, d).
        y: `jax.Array`
            The second input data of shape = (n_y, d).

        Returns
        -------
        jax.Array:
            The model matrix for the dynamice time warping measure of shape (n_x + n_y - 1, n_y).
        """
        x = jnp.expand_dims(x, axis=1)
        y = jnp.expand_dims(y, axis=1)
        distance_matrix = jax.vmap(
            jax.vmap(self.distance, in_axes=(0, None)), in_axes=(None, 0)
        )(x, y)

        h, _ = distance_matrix.shape
        rows = []
        for row in range(h):
            rows.append(
                jnp.pad(
                    distance_matrix[row], (row, h - row - 1), constant_values=jnp.inf
                )
            )
        return jnp.stack(rows, axis=1)

    def run(self, x: jax.Array, y: jax.Array) -> jax.Array:
        """
        Estimate the Dynamic Time Warping distance measure.

        Parameters
        ----------
        x: `jax.Array`
            The input data point of shape (d, ) if particle, or (n_x, d) if time series.
        y: `jax.Array`
            The input data point of shape (d, ) if particle, or (n_y, d) if time series.

        Returns
        -------
        `jax.Array`:
            The estimated Dynamic Time Warping distance of shape ().
        """
        assert (
            x.shape[-1] == y.shape[-1]
        ), f"The two inputs need to have the same dimensionality. Got x = {x.shape} and y = {y.shape}."
        assert x.ndim <= 2 and y.ndim <= 2, (
            f"The two inputs need to be of shape (d, ) if particle, or (n, d) if time series. "
            f"Got x = {x.shape} and y = {y.shape}."
        )

        if x.ndim == 1:
            x = jnp.expand_dims(x, axis=0)
        if y.ndim == 1:
            y = jnp.expand_dims(y, axis=0)

        def _body_fn(carry: Sequence, anti_diagonal: jax.Array) -> Any:
            two_ago, one_ago = carry

            diagonal = two_ago[:-1]
            right = one_ago[:-1]
            down = one_ago[1:]
            best = jnp.min(jnp.stack([diagonal, right, down], axis=-1), axis=-1)

            next_row = best + anti_diagonal
            next_row = jnp.pad(next_row, (1, 0), constant_values=jnp.inf)

            return (one_ago, next_row), next_row

        model_matrix = self.init_model_matrix(x, y)

        init = (
            jnp.pad(model_matrix[0], (1, 0), constant_values=jnp.inf),
            jnp.pad(
                model_matrix[1] + model_matrix[0, 0], (1, 0), constant_values=jnp.inf
            ),
        )
        carry, ys = jax.lax.scan(_body_fn, init, model_matrix[2:], unroll=2)
        return carry[1][-1]


# --------------------------------------------------------------------------------------------------------------------
# --------------------------------------------- Discrete Frechet Distance --------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class DiscreteFrechetDistance(DistanceMeasures):
    """
    Similarity measure between time series data using the discrete Frechet distance [1. 2]. It is the minimum length of
    a leash required for a dog and its owner to walk along their respective curves without backtracking. This implies
    that the series can be of different lengths. The discrete Frechet distance is a measure of similarity between two
    curves.

    The algorithm follows a dynamic programming approach to find the optimal alignment between the two sequences. We
    followed the implementation of DTW from [3] and replaced the sum operation with the max operation resulting in
    the Discrete Frechet distance.

    Parameters
    ----------
    distance: `DistanceMeasures`, optional, default = None.
        The distance measure to use for the discrete Frechet distance. If None, the Euclidean distance is used.

    Returns
    -------
    `DiscreteFrechetDistance`
        An instance of the discrete Frechet distance measure.

    References
    ----------
    [1] M. Fr ́echet. Sur quelques points du calcul fonctionnel. 1906.
      Available: https://link.springer.com/article/10.1007/BF03018603
    [2] T. Eiter and H. Mannila. Computing discrete fr ́echet distance. 1994.
      Available: http://www.kr.tuwien.ac.at/staff/eiter/et-archive/cdtr9464.pdf
    [3] K. Heidler. (Soft-)DTW for JAX, Github, https://github.com/khdlr/softdtw_jax
    """

    distance: Optional[DistanceMeasures] = struct.field(default=None, pytree_node=False)

    @classmethod
    def construct(
        cls, distance: Optional[DistanceMeasures] = None
    ) -> "DiscreteFrechetDistance":
        """
        Construct the discrete Frechet distance measure.

        Parameters
        ----------
        distance: `DistanceMeasures`, optional, default = None.
            The distance measure to use for the discrete Frechet distance. If None, the Euclidean distance is used.

        Returns
        -------
        `DiscreteFrechetDistance`
            An instance of the discrete Frechet distance measure.
        """
        if distance is None:
            distance = EuclideanDistance.construct()
        return cls(distance=distance)

    def init_model_matrix(self, x: jax.Array, y: jax.Array) -> jax.Array:
        """
        Initialize the state for the discrete Frechet distance measure.

        Parameters
        ----------
        x: `jax.Array`
            The input data of shape = (n_x, d).
        y: `jax.Array`
            The second input data of shape = (n_y, d).

        Returns
        -------
        jax.Array:
            The model matrix for the discrete Frechet distance measure of shape (n_x + n_y - 1, n_y).
        """
        x = jnp.expand_dims(x, axis=1)
        y = jnp.expand_dims(y, axis=1)
        distance_matrix = jax.vmap(
            jax.vmap(self.distance, in_axes=(0, None)), in_axes=(None, 0)
        )(x, y)

        h, _ = distance_matrix.shape

        rows = []
        for row in range(h):
            rows.append(
                jnp.pad(
                    distance_matrix[row], (row, h - row - 1), constant_values=jnp.inf
                )
            )
        return jnp.stack(rows, axis=1)

    def run(self, x: jax.Array, y: jax.Array) -> jax.Array:
        """
        Estimate the discrete Frechet distance measure.

        Parameters
        ----------
        x: `jax.Array`
            The input data point of shape (d, ) if particle, or (n_x, d) if time series.
        y: `jax.Array`
            The input data point of shape (d, ) if particle, or (n_y, d) if time series.

        Returns
        -------
        `jax.Array`:
            The estimated Dynamic Time Warping distance of shape ().
        """
        assert (
            x.shape[-1] == y.shape[-1]
        ), f"The two inputs need to have the same dimensionality. Got x = {x.shape} and y = {y.shape}."
        assert x.ndim <= 2 and y.ndim <= 2, (
            f"The two inputs need to be of shape (d, ) if particle, or (n, d) if time series. "
            f"Got x = {x.shape} and y = {y.shape}."
        )

        if x.ndim == 1:
            x = jnp.expand_dims(x, axis=0)
        if y.ndim == 1:
            y = jnp.expand_dims(y, axis=0)

        def _body_fn(carry: Sequence, anti_diagonal: jax.Array) -> Any:
            two_ago, one_ago = carry

            diagonal = two_ago[:-1]
            right = one_ago[:-1]
            down = one_ago[1:]
            best = jnp.min(jnp.stack([diagonal, right, down], axis=-1), axis=-1)

            next_row = jnp.maximum(best, anti_diagonal)
            next_row = jnp.pad(next_row, (1, 0), constant_values=jnp.inf)

            return (one_ago, next_row), next_row

        model_matrix = self.init_model_matrix(x, y)

        init = (
            jnp.pad(model_matrix[0], (1, 0), constant_values=jnp.inf),
            jnp.pad(
                jnp.maximum(model_matrix[1], model_matrix[0]),
                (1, 0),
                constant_values=jnp.inf,
            ),
        )

        carry, ys = jax.lax.scan(_body_fn, init, model_matrix[2:], unroll=2)
        return carry[1][-1]


# --------------------------------------------------------------------------------------------------------------------
# ------------------------------------------------- Sinkhorn Distance ------------------------------------------------
# --------------------------------------------------------------------------------------------------------------------
@struct.dataclass
class OTTCostWrapper(CostFn):
    """
    The base cost function for the Sinkhorn distance measure. The cost function is a weighted sum of the spatial and
    temporal distances.

    Parameters
    ----------
    weights: `Sequence[float]`, optional, default = None.
        The weights for the spatial and temporal distances. If None, the weights are set to [1., 1.].
    distances: `Sequence[DistanceMeasures]`, optional, default = None.
        The spatial and temporal distance measures. If None, the spatial distance is the squared Euclidean distance

    Returns
    -------
    `OTTCostWrapper`
        An instance of the base cost function.
    """
    weights: Optional[Sequence[float]] = None
    distances: Optional[Sequence[DistanceMeasures]] = None

    @classmethod
    def construct(cls, *args: Any, **kwargs: Any) -> "OTTCostWrapper":
        """
        Construct the base cost function for the Sinkhorn distance measure.

        Parameters
        ----------
        args: `Any`
            The arguments to pass to the constructor.
        kwargs: `Any`
            The keyword arguments to pass to the constructor.

        Returns
        -------
        `OTTCostWrapper`
            An instance of the base cost function.
        """
        weights = [1.0, 1.0]
        distances = [
            SquaredEuclideanDistance.construct(),
            MinkowskiDistance.construct(p=1),
        ]
        return cls(weights=weights, distances=distances)

    @classmethod
    def create(cls, *args: Any, **kwargs: Any) -> "OTTCostWrapper":
        """
        Create an instance of the base cost function for the Sinkhorn distance measure.

        Parameters
        ----------
        args: `Any`
            The arguments to pass to the constructor.
        kwargs: `Any`
            The keyword arguments to pass to the constructor.

        Returns
        -------
        `OTTCostWrapper`
            An instance of the base cost function.
        """
        return cls.construct(*args, **kwargs)

    def __call__(self, x: jnp.ndarray, y: jnp.ndarray) -> float:
        """
        Calculate the pairwise cost between two points.

        Parameters
        ----------
        x: `jnp.ndarray`
            The first point of shape (d, ).
        y: `jnp.ndarray`
            The second point of shape (d, ).

        Returns
        -------
        `float`
            The pairwise cost between the two points.
        """
        total_cost = 0
        spatio_weights, temporal_weights = self.weights
        spatio_dist, temporal_dist = self.distances

        # Spatial coordinate related cost
        total_cost += spatio_weights * spatio_dist(x[:-1], y[:-1]).squeeze()

        # Temporal coordinate related cost
        total_cost += temporal_weights * temporal_dist(x[-1:], y[-1:]).squeeze()
        return total_cost


@struct.dataclass
class SinkhornDistance(DistanceMeasures):
    """
    Similarity measure between data points using the Sinkhorn distance measure [1, 2]. The Sinkhorn distance is a
    regularized optimal transport distance that is a measure of similarity between two probability distributions. Here,
    we consider each point as a weighted particle.

    The algorithm follows a linear programming approach to find the optimal transport plan between the two sequences.
    We used the implementation of the Sinkhorn distance from [3].

    Parameters
    ----------
    solver: `Any`, optional, default = None.
        The solver to use for the Sinkhorn distance. If None, the Sinkhorn solver is used.
    cost_fn: `CostFn`, optional, default = None.
        The cost function to use for the Sinkhorn distance. If None, the base cost function is used.
    epsilon: `float`, optional, default = None.
        The regularization parameter for the Sinkhorn distance. If None, the default value is used.
    return_regularized_cost: `bool`, optional, default = False.
        If True, the regularized cost is returned. If False, the unregularized cost is returned.

    Returns
    -------
    `SinkhornDistance`
        An instance of the Sinkhorn distance measure.

    References
    ----------
    [1] M. Cuturi. Sinkhorn distances: Lightspeed computation of optimal transport. 2013.
        Available: https://arxiv.org/abs/1306.0895
    [2] G. Peyré and M. Cuturi. Computational Optimal Transport. 2019.
        Available: https://arxiv.org/abs/1803.00567
    [3] M. Cuturi. Optimal Transport Tools (OTT): A JAX Toolbox for all things Wasserstein, arXiv, 2022.
        Available: arXiv preprint arXiv:2201.12324
        Github: https://github.com/ott-jax/ott
        Docs: https://ott-jax.readthedocs.io/en/latest/
    """

    solver: Optional[Any] = None
    cost_fn: Optional[CostFn] = struct.field(default=None, pytree_node=False)
    epsilon: Optional[float] = struct.field(default=None, pytree_node=False)
    return_regularized_cost: bool = struct.field(default=False, pytree_node=False)

    @classmethod
    def construct(
        cls,
        epsilon: Optional[float] = None,
        cost_fn: Optional[CostFn] = None,
        return_regularized_cost: bool = False,
    ) -> "SinkhornDistance":
        """
        Construct the Sinkhorn distance measure.

        Parameters
        ----------
        epsilon: `float`, optional, default = None.
            The regularization parameter for the Sinkhorn distance.
        cost_fn: `CostFn`, optional, default = None.
            The cost function to use for the Sinkhorn distance.
        return_regularized_cost: `bool`, optional, default = None.
            If True, the regularized cost is returned. If False, the unregularized cost is returned.

        Returns
        -------
        `SinkhornDistance`
            An instance of the Sinkhorn distance measure.
        """
        if cost_fn is None:
            cost_fn = OTTCostWrapper.construct()

        return cls(
            solver=sinkhorn.Sinkhorn(),
            epsilon=epsilon,
            cost_fn=cost_fn,
            return_regularized_cost=return_regularized_cost,
        )

    def init_geometry(self, x: jax.Array, y: jax.Array) -> Any:
        """
        Initialize the geometry for the Sinkhorn distance measure.

        Parameters
        ----------
        x: `jax.Array`
            The input data of shape = (n_x, d).
        y: `jax.Array`
            The second input data of shape = (n_y, d).

        Returns
        -------
        `Any`
            The geometry of the Sinkhorn distance measure.
        """
        # Add time to given arrays based on a linear interpolation
        n_x, _ = x.shape
        x_extended = jnp.concatenate(
            (x, jnp.linspace(0, 1, n_x)[:, jnp.newaxis]), axis=-1
        )
        n_y, _ = y.shape
        y_extended = jnp.concatenate(
            (y, jnp.linspace(0, 1, n_y)[:, jnp.newaxis]), axis=-1
        )

        # Generate and return a geometry for a Linear OT problem
        geometry = pointcloud.PointCloud(
            x_extended, y_extended, cost_fn=self.cost_fn, epsilon=self.epsilon
        )
        return geometry

    def run(self, x: jax.Array, y: jax.Array) -> jax.Array:
        """
        Estimate the Sinkhorn distance measure.

        Parameters
        ----------
        x: `jax.Array`
            The input data point of shape (d, ) if particle, or (n_x, d) if time series.
        y: `jax.Array`
            The input data point of shape (d, ) if particle, or (n_y, d) if time series.

        Returns
        -------
        `jax.Array`
            The estimated Sinkhorn distance of shape ().
        """
        assert (
            x.shape[-1] == y.shape[-1]
        ), f"The two inputs need to have the same dimensionality. Got x = { x.shape} and y = {y.shape}."
        assert x.ndim <= 2 and y.ndim <= 2, (
            f"The two inputs need to be of shape (d, ) if particle, or (n, d) if time series. "
            f"Got x = {x.shape} and y = {y.shape}."
        )

        if x.ndim == 1:
            x = jnp.expand_dims(x, axis=0)
        if y.ndim == 1:
            y = jnp.expand_dims(y, axis=0)

        geometry = self.init_geometry(x, y)
        ot_problem = linear_problem.LinearProblem(geometry)
        solution = self.solver(ot_problem)
        if self.return_regularized_cost:
            return solution.reg_ot_cost
        return jnp.sum(solution.matrix * solution.geom.cost_matrix)
